'''
This file implements the Binary Search Tree data structure.
The functions in this file are considerably harder than the
functions in the BinaryTree file.
'''

from containers.BinaryTree import BinaryTree, Node


class BST(BinaryTree):
    '''
    The BST is a superclass of BinaryTree.
    That means that the BST class "inherits" all of the methods
    from BinaryTree,
    and we don't have to reimplement them.
    '''

    def __init__(self, xs=None):
        '''
        FIXME:
        If xs is a list (i.e. xs is not None),
        then each element of xs needs to be inserted into the BST.
        '''
        super().__init__()
        if xs is not None:
            for x in xs:
                self.insert(x)

    def __repr__(self):
        '''
        Notice that in the BinaryTree class,
        we defined a __str__ function,
        but not a __repr__ function.
        Recall that the __repr__ function should return a string that
        can be used to recreate a valid instance of the class.
        Thus, if you create a variable using the command BST([1,2,3])
        it's __repr__ will return "BST([1,2,3])"

        For the BST, type(self).__name__ will be the string "BST",
        but for the AVLTree, this expression will be "AVLTree".
        Using this expression ensures that all subclasses of BST
        will have a correct implementation of __repr__,
        and that they won't have to reimplement it.
        '''
        return type(self).__name__ + '(' + str(self.to_list('inorder')) + ')'

    def __eq__(self, t2):
        '''
        This method checks to see if the contents of self and t2 are equal.
        The expression `a == b` desugars to `a.__eq__(b)`.

        NOTE:
        We only care about "semantic" equality,
        and not "syntactic" equality.
        That is, we do not care about the tree structure itself,
        and only care about the contents of what the tree contains.

        HINT:
        Convert the contents of both trees into a sorted list,
        then compare those sorted lists for equality.
        '''
        list_t1 = self.to_list('inorder')
        list_t2 = t2.to_list('inorder')
        for x in list_t1:
            if x not in list_t2:
                return False
        for x in list_t2:
            if x not in list_t1:
                return False
        return True

    def is_bst_satisfied(self):
        '''
        Whenever you implement a data structure,
        the first thing to do is to implement a function that checks whether
        the structure obeys all of its laws.
        This makes it possible to automatically test whether
        insert/delete functions
        are actually working.
        '''
        if self.root:
            return BST._is_bst_satisfied(self.root)
        return True

    @staticmethod
    def _is_bst_satisfied(node):
        '''
        FIXME:
        The current implementation has a bug:
        it only checks if the children of the current node are
        less than/greater than, rather than ensuring that all nodes
        to the left/right are less than/greater than.

        HINT:
        Use the _find_smallest and _find_largest functions to fix the bug.
        You should use the _ prefixed methods because those are static
        methods just like this one.
        '''
        ret = True
        if node is None:
            return True
        if node.left:
            if node.value >= BST._find_largest(node.left):
                ret &= BST._is_bst_satisfied(node.left)
            else:
                ret = False
        if node.right:
            if node.value <= BST._find_smallest(node.right):
                ret &= BST._is_bst_satisfied(node.right)
            else:
                ret = False
        return ret

    def insert(self, value):
        '''
        Inserts value into the BST.

        FIXME:
        Implement this function.

        HINT:
        Create a staticmethod helper function following the pattern
        of _is_bst_satisfied.
        '''
        if not self.root:
            self.root = Node(value)
            return
        if value == self.root.value:
            return
        else:
            return self._insert(value, self.root)

    @staticmethod
    def _insert(value, node):
        '''
        FIXME:
        Implement this function.
        '''
        if node.value == value:
            return
        if value < node.value:
            if node.left is None:
                node.left = Node(value)
                return
            else:
                return BST._insert(value, node.left)
        else:
            if node.right is None:
                node.right = Node(value)
            else:
                return BST._insert(value, node.right)

    def insert_list(self, xs):
        '''
        Given a list xs, insert each element of xs into self.

        FIXME:
        Implement this function.

        HINT:
        Repeatedly call the insert method.
        You cannot get this method to work correctly until you have gotten
        insert to work correctly.
        '''
        for x in xs:
            self.insert(x)

    def __contains__(self, value):
        '''
        Recall that `x in tree` desugars to `tree.__contains__(x)`.
        '''
        return self.find(value)

    def find(self, value):
        '''
        Returns whether value is contained in the BST.

        FIXME:
        Implement this function.
        '''
        if not self.root:
            return False
        if value == self.root.value:
            return True
        else:
            if value < self.root.value:
                return self._find(value, self.root.left)
            else:
                return self._find(value, self.root.right)

    @staticmethod
    def _find(value, node):
        '''
        FIXME:
        Implement this function.
        '''
        if node is None:
            return False
        if node.value == value:
            return True
        if value < node.value:
            return BST._find(value, node.left)
        else:
            return BST._find(value, node.right)

    def find_smallest(self):
        '''
        Returns the smallest value in the tree.
        '''
        if self.root is None:
            raise ValueError('Nothing in tree')
        else:
            return BST._find_smallest(self.root)

    @staticmethod
    def _find_smallest(node):
        '''
        This is a helper function for find_smallest and not intended
        to be called directly by the user.
        '''
        assert node is not None
        if node.left is None:
            return node.value
        else:
            return BST._find_smallest(node.left)

    def find_largest(self):
        '''
        Returns the smallest value in the tree.
        '''
        if self.root is None:
            raise ValueError('Nothing in tree')
        else:
            return BST._find_largest(self.root)

    @staticmethod
    def _find_largest(node):
        '''
        Returns the largest value in the tree.

        FIXME:
        Implement this function.

        HINT:
        Follow the pattern of the _find_smallest function.
        '''
        assert node is not None
        if node.right is None:
            return node.value
        else:
            return BST._find_largest(node.right)

    def remove(self, value):
        '''
        Removes value from the BST.
        If value is not in the BST, it does nothing.

        FIXME:
        Implement this function.

        HINT:
        You should have everything else working before you implement
        this function.

        HINT:
        Use a recursive helper function.
        '''
        if not self.root:
            return self.root
        else:
            self.root = self._remove(value, self.root)

    @staticmethod
    def _remove(value, node):
        '''
        FIXME:
        Implement this function.
        '''
        if node is None:
            return node
        if value < node.value:
            node.left = BST._remove(value, node.left)
            return node
        elif value > node.value:
            node.right = BST._remove(value, node.right)
            return node
        if node.value == value:
            if node.left is None:
                return node.right
            elif node.right is None:
                return node.left
            else:
                min_of_right = BST._find_smallest(node.right)
                node.value = min_of_right
                node.right = BST._remove(min_of_right, node.right)
                return node

    def remove_list(self, xs):
        '''
        Given a list xs, remove each element of xs from self.
        FIXME:
        Implement this function.

        HINT:
        See the insert_list function.
        '''
        for x in xs:
            self.remove(x)



import unicodedata


class NormalizedStr:
    '''
    By default, Python's str type stores any valid unicode string.
    This can result in unintuitive behavior.
    For example:

    >>> 'César' in 'César Chávez'
    True
    >>> 'César' in 'César Chávez'
    False

    The two strings to the right of the in keyword above are equal *semantically*,
    but not equal *representationally*.
    In particular, the first is in NFC form, and the second is in NFD form.
    The purpose of this class is to automatically normalize our strings for us,
    making foreign languages "just work" a little bit easier.
    '''

    def __init__(self, text, normal_form='NFC'):
        self.nf = normal_form
        self.text = unicodedata.normalize(self.nf, text)
        self.i = -1

    def __repr__(self):
        '''
        The string returned by the __repr__ function should be valid python code
        that can be substituted directly into the python interpreter to reproduce an equivalent object.
        '''
        return "NormalizedStr(\'{}\', \'{}\')".format(self.text, self.nf)

    def __str__(self):
        '''
        This functions converts the NormalizedStr into a regular string object.
        The output is similar, but not exactly the same, as the __repr__ function.
        '''
        return str(self.text)

    def __len__(self):
        '''
        Returns the length of the string.
        The expression `len(a)` desugars to a.__len__().
        '''
        return len(self.text)

    def __contains__(self, substr):
        '''
        Returns true if the `substr` variable is contained within `self`.
        The expression `a in b` desugars to `b.__contains__(a)`.

        HINT:
        You should normalize the `substr` variable to ensure that the comparison is done semantically and not syntactically.
        '''
        norm_ss = unicodedata.normalize(self.nf, substr)
        return str(self.text).__contains__(str(norm_ss))

    def __getitem__(self, index):
        '''
        Returns the character at position `index`.
        The expression `a[b]` desugars to `a.__getitem__(b)`.
        '''
        return self.text.__getitem__(index)

    def lower(self):
        '''
        Returns a copy in the same normalized form, but lower case.
        '''
        return self.text.lower()

    def upper(self):
        '''
        Returns a copy in the same normalized form, but upper case.
        '''
        return self.text.upper()

    def __add__(self, b):
        '''
        Returns a copy of `self` with `b` appended to the end.
        The expression `a + b` gets desugared into `a.__add__(b)`.

        HInT:
        The addition of two normalized strings is not guaranteed to stay normalized.
        Therefore, you must renormalize the strings after adding them together.
        '''
        new_text = unicodedata.normalize(self.nf, self.text + unicodedata.normalize(self.nf, str(b)))
        return NormalizedStr(new_text, self.nf)

    def __iter__(self):
        '''
        HINT:
        Recall that the __iter__ method returns a class, which is the iterator object.
        You'll need to define your own iterator class with the appropriate magic methods,
        and return an instance of that class here.
        '''
        return self
        # return NormStrIter(self.text)

    def __next__(self):
        if self.i == len(self.text) - 1:
            raise StopIteration
        else:
            self.i += 1
            return self.text[self.i]


class NormStrIter:
    def __init__(self, text):
        self.text = text
        self.i = -1

    def __next__(self):
        if self.i == len(self.text) - 1:
            raise StopIteration
        else:
            self.i += 1
            return self.text[self.i]




###############################################################################
# example fibonacci number code;
# you do not have to modify this code in any way
###############################################################################


def fibs(n):
    '''
    This function computes the first n fibonacci numbers.
    Notice that this function uses O(n) memory.
    '''
    fibs = []
    fibs.append(1)
    if n == 1:
        return fibs
    fibs.append(1)
    while len(fibs) < n:
        fibs.append(fibs[-1] + fibs[-2])
    return fibs


def fib_bad(n):
    '''
    This function computes the n-th fibonacci number,
    but it uses O(n) memory to do so,
    which is bad.
    '''
    return fibs(n)[-1]


def fib(n):
    '''
    This function computes the n-th fibonacci number,
    but it consumes only O(1) memory,
    and is optimal.
    '''
    if n < 2:
        return 1
    f0 = 1
    f1 = 1
    for i in range(n - 1):
        f2 = f1 + f0
        f0 = f1
        f1 = f2
    return f2


###############################################################################
# fibonacci number code using generators;
# you will need to implement the functions below
###############################################################################


class Fib:
    '''
    This class represents all the fibonacci numbers,
    but uses O(1) memory to do so.

    >>> list(Fib(5))
    [1, 1, 2, 3, 5]
    '''
    def __init__(self, n=None):
        self.n = n

    def __iter__(self):
        if self.n is not None:
            return FibIter(self.n - 1)
        else:
            return FibIter()

    def __repr__(self):
        if self.n is not None:
            return f"Fib({self.n})"
        else:
            return 'Fib()'


class FibIter:
    '''
    This is the iterator helper class for the Fib class.
    '''
    def __init__(self, n=None):
        self.n = n
        self.f0 = 1
        self.f1 = 1
        self.ret = 0
        self.i = 0

    def __next__(self):
        if self.n is not None and self.n + 1 <= self.i:
            raise StopIteration
        else:
            tmp = self.f0
            self.f2 = self.f1 + self.f0
            self.f0 = self.f1
            self.f1 = self.f2
            self.i += 1
            return tmp


def fib_yield(n=None):
    '''
    This function returns a generator that computes the
    first n fibonacci numbers.
    If n is None, then the generator is infinite.
    '''
    if n is not None and n <= 2:
        for _ in range(n):
            yield 1
    elif n is None:
        for _ in range(2):
            yield 1
        while True:
            f0 = 0
            f1 = 1
            i = 1
            f2 = f1 + f0
            f0 = f1
            f1 = f2
            yield f0
    else:
        f0 = 0
        f1 = 1
        i = 1
        for i in range(n):
            f2 = f1 + f0
            f0 = f1
            f1 = f2
            yield f0



def range(a, b=None, c=None):
    '''
    This function should behave exactly like the built-in range function.
    For example:

    >>> list(range(5))
    [0, 1, 2, 3, 4]
    >>> list(range(1, 5))
    [1, 2, 3, 4]
    >>> list(range(1, 5, 2))
    [1, 3]

    HINT:
    If you can figure out how to use the built-in range function
    (without modifying the test cases!),
    then feel free to do so.
    That's fairly difficult to do, however, and it's much easier to just
    implement this function normally using the yield syntax.

    NOTE:
    For efficiency reasons, Python's built-in range object is written in
    the C programming language rather than natively in python.
    You can find the source code online at
    https://hg.python.org/cpython/file/ee7b713fec71/Objects/rangeobject.c
    The link takes you to a file that is 1445 lines long,
    and all it does is implement this simple functionality.
    My easy to read Python implementation of this function is
    just 15 lines long.
    (And with some code golf I also wrote a less readable version
    that is only 4 lines.)
    It is very common for C programs to be 100x longer than their
    corresponding python programs.
    C code must manage lots of details about the computer manually
    that python code automates for you.
    Carefully written C code can be faster than the corresponding
    python code because it can remove some of the overhead of this
    automation process,
    but the resulting code is much longer and harder to read/write.
    '''
    if c is None:
        c = 1
    if b is None:
        b = a
        a = 0
    reverse = False
    if a > b:
        reverse = True
    i = a
    if (b - a) / c < 0:
        return []
    if not reverse:
        while i < b:
            yield i
            i += c
    else:
        while i > b:
            yield i
            i += c



'''
This file implements the Heap data structure as a subclass of the BinaryTree.
The book implements Heaps using an *implicit* tree with an *explicit* vector implementation,
so the code in the book is likely to be less helpful than the code for the other data structures.
The book's implementation is the traditional implementation because it has a faster constant factor
(but the same asymptotics).
'''

from containers.BinaryTree import BinaryTree


class Heap(BinaryTree):
    '''
    FIXME:
    Heap is currently not a subclass of BinaryTree.
    You should make the necessary changes in the class declaration line above
    and in the constructor below.
    '''

    def __init__(self, xs=None):
        '''
        FIXME:
        If xs is a list (i.e. xs is not None),
        then each element of xs needs to be inserted into the Heap.
        '''
        super().__init__()
        self.numel = 0
        self.tree_list = [float('-inf')]
        if xs is not None:
            self.insert_list(list(xs))

    def __repr__(self):
        '''
        Notice that in the BinaryTree class,
        we defined a __str__ function,
        but not a __repr__ function.
        Recall that the __repr__ function should return a string that can be used to recreate a valid instance of the class.
        Thus, if you create a variable using the command Heap([1,2,3])
        it's __repr__ will return "Heap([1,2,3])"

        For the Heap, type(self).__name__ will be the string "Heap",
        but for the AVLTree, this expression will be "AVLTree".
        Using this expression ensures that all subclasses of Heap will have a correct implementation of __repr__,
        and that they won't have to reimplement it.
        '''
        return type(self).__name__ + '(' + str(self.to_list('inorder')[1:]) + ')'

    def is_heap_satisfied(self):
        '''
        Whenever you implement a data structure,
        the first thing to do is to implement a function that checks whether
        the structure obeys all of its laws.
        This makes it possible to automatically test whether insert/delete functions
        are actually working.
        '''
        if self.root:
            return Heap._is_heap_satisfied(self.root)
        return True

    @staticmethod
    def _is_heap_satisfied(node):
        '''
        FIXME:
        Implement this method.
        '''
        if node is None:
            return True
        ret = True
        if node.left:
            if node.left.value < node.value:
                return False
            else:
                ret &= Heap._is_heap_satisfied(node.left)
        if node.right:
            if node.right.value < node.value:
                return False
            else:
                ret &= Heap._is_heap_satisfied(node.right)
        return ret

    def insert(self, value):
        '''
        Inserts value into the heap.

        FIXME:
        Implement this function.

        HINT:
        The pseudo code is
        1. Find the next position in the tree using the binary representation of the total number of nodes
            1. You will have to explicitly store the size of your heap in a variable (rather than compute it) to maintain the O(log n) runtime
            1. See https://stackoverflow.com/questions/18241192/implement-heap-using-a-binary-tree for hints
        1. Add `value` into the next position
        1. Recursively swap value with its parent until the heap property is satisfied

        HINT:
        Create a @staticmethod helper function,
        following the same pattern used in the BST and AVLTree insert functions.
        '''
        self.numel += 1
        binary = "{0:b}".format(self.numel)[1:]
        self.tree_list.append(value)
        self.do_swaps(binary)

    def do_swaps(self, binary):
        if binary == '':
            return
        if not self.is_heap_satisfied():
            idx = int(binary)
            idx_2 = int(binary[:-1])
            if self.tree_list[idx] < self.tree_list[idx_2]:
                tmp = self.tree_list[idx]
                self.tree_list[idx] = self.tree_list[idx_2]
                self.tree_list[idx_2] = tmp
        if not self.is_heap_satisfied():
            self.do_swaps(binary[:-1])

    def do_down_swaps(self, binary):
        if not self.is_heap_satisfied():
            idx = int(binary)
            while (idx * 2) <= self.numel:
                mc = self.min_child(idx)
                if self.tree_list[idx] > self.tree_list[mc]:
                    tmp = self.heap_tree[idx]
                    self.heap_tree[idx] = self.heap_tree[mc]
                    self.heap_tree[mc] = tmp
                idx = mc

    def min_child(self, idx):
        if idx * 2 + 1 > self.numel:
            return idx * 2
        else:
            if self.tree_list[idx * 2] < self.tree_list[idx * 2 + 1]:
                return idx * 2
            else:
                return idx * 2 + 1

    def to_list(self, order):
        return self.tree_list

    def insert_list(self, xs):
        '''
        Given a list xs, insert each element of xs into self.

        FIXME:
        Implement this function.
        '''
        for x in xs:
            self.insert(x)

    def find_smallest(self):
        '''
        Returns the smallest value in the tree.

        FIXME:
        Implement this function.
        '''
        return min(self.tree_list[1:])

    def remove_min(self):
        '''
        Removes the minimum value from the Heap.
        If the heap is empty, it does nothing.

        FIXME:
        Implement this function.

        HINT:
        The pseudocode is
        1. remove the bottom right node from the tree
        2. replace the root node with what was formerly the bottom right
        3. "trickle down" the root node: recursively swap it with its largest child until the heap property is satisfied

        HINT:
        I created two @staticmethod helper functions: _remove_bottom_right and _trickle.
        It's possible to do it with only a single helper (or no helper at all),
        but I personally found dividing up the code into two made the most sense.
        '''
        retval = self.find_smallest()
        idx = self.tree_list.index(retval)
        self.tree_list[idx] = self.tree_list[self.numel]
        self.numel -= 1
        self.tree_list.pop()
        self.do_down_swaps(idx)
        return retval
